"""
Rollet settings and its methods
"""

__all__ = [
    "FULLTEXT",
    "ABSTRACT",
    "TITLE"
]

from rollet.extractor import *

import os 
dir_path = os.path.dirname(os.path.realpath(__file__))


# CSS rules selectors for each field
FULLTEXT = (
    ':is(span,div)[class="contenuto" i]',
    'div[class*="full" i][class*="text" i] a',
    *[f'a[{attr}*="full" i][{attr}*="text" i]' for attr in ('class', 'data-ga-category')],
    *[f':is(div,section)[{attr}*="sec-" i]' for attr in ('class', 'id')],
    ':is(div,section,article)[class*="article" i][class*="section" i][class*="content" i]',
    ':is(div,section,article)[class*="full" i][class*="text" i]',
    'a[class*="pdf" i]',
    *[f'div[{attr}*="download" i] a' for attr in ('class', 'label')],
    'div[id^="Sec"]',
    *[f'div[{attr}*="{value}" i]' for attr in ('class', 'id') for value in ('article', 'content')],
    *[f':is(article,div)[{attr}*="body" i]' for attr in ('id', 'class', 'itemprop')]
)

ABSTRACT = (
    *[f':is(div,span)[class*="{value}" i]' for value in ('contenuto', 'sommario')],
    *[f':is(article,section,div)[{attr}*="blog" i][{attr}*="post" i]' for attr in ('class', 'id')],
    *[f':is(div,p)[class*="{value}" i]' for value in ('news-txt', 'desc', 'intro')],
    *[f'meta[name="{attr}" i]'
    for attr in ('description', 'twitter:description', 'citation_abstract')],
    'script[type*="json" i]',
    'meta[name="dc.description" i]',
    'meta[property="og:description" i]',
    *[f':is(div,section)[{attr}*="abs" i]' for attr in ('class', 'id')],
    'div[class*=content i]',
    *[f':is(div,section)[{attr}*="{value}" i]' for attr in ('class', 'id')
    for value in ('-abstract', 'abstract')],
    *[f'div[{attr}*="article" i][{attr}*="body" i]'
    for attr in ('class', 'itemprop')],
)

TITLE = (
    'meta[property="og:title" i]',
    *[f'meta[name="{attr}" i]'
    for attr in ('citation_title', 'dc.title', 'twitter:title', 'title')],
    'script[type*="json" i]',
    *[f':is(h1,h2,h3,div,span)[{attr}*=title i]'
    for attr in ('class', 'id', 'itemprop')],
    *[f':is(h1,h2,h3)[itemprop*=head i]'],
    'title'
)

# Domains types
types = {
    'article': [
        'agronews', 'ansa', 'berkeley', 'citrusindustry', 'europa', 'europafm',
        'europalibera', 'europapress', 'europatat', 'floraldaily',
        'freshfruitportal', 'freshplaza', 'groentennieuws', 'hortidaily',
        'hortweek', 'huffingtonpost', 'ica', 'mon-viti', 'nieuweoogst', 'nvwa',
        'oliveoiltimes', 'phytoma', 'phytoma-ldv', 'teatronaturale',
        'terrenature', 'theguardian', 'usda', 'vniikr'
    ],
    'paper': [
        'apsnet', 'biorxiv', 'chinapubmed', 'europepmc', 'mdpi', 'nature',
        'nih', 'nl', 'researchgate', 'sciencedirect', 'springer',
        'srpingeropen', 'wiley'
    ]
}
TYPES = dict(sum([[(i,k) for i in v] for k,v in types.items()], []))

# Types defaults keywords arguments for fetch
FETCH_KWARGS = {
    'default': {
        'fulltext__which': 'first',
        'fulltext__script_keys': [],
        'abstract__which': 'max',
        'abstract__script_keys': [],
        'lang__attr': 'lang',
        'title__which': 'first',
        'title__script_keys': []
    },'article': {
        'fulltext__which': 'max',
        'fulltext__script_keys': [],
        'abstract__which': 'max',
        'abstract__script_keys': ['abstract', 'summary'],
        'lang__attr': 'lang',
        'title__which': 'first',
        'title__script_keys': ['title', 'heading']
    },
    'paper': {
        'fulltext__which': 'max',
        'fulltext__script_keys': [],
        'abstract__which': 'first',
        'abstract__script_keys': ['abstract', 'summary'],
        'lang__attr': 'lang',
        'title__which': 'first',
        'title__script_keys': ['title', 'heading', 'name']
    },
}

# Blacklisted domains
with open(os.path.join(dir_path, 'blacklist'), errors = 'ignore') as f:
    BLACKLIST = list(map(lambda x:x.strip(), f.readlines()))

# Custom domains extractors
# {domain: [(schema1, Extractor1), (schema2, Extractor2)]}
EXTRACTORS = {
    'default': [(r'.*', BaseExtractor),],
    'mdpi': [(r'.*', MdpiExtractor),],
    'nih': [
        (r'/pmc/articles/', PMCExtractor),
        (r'.*', NihExtractor),],
    'sciencedirect': [(r'.*', ScienceDirectExtractor),],
    'springer': [(r'.*', SpringerExtractor),],
    'apsnet': [(r'.*', ApsNetExtractor),],
    'wiley': [(r'.*', WileyExtractor),],
    'agronews': [(r'.*', AgronewsExtractor),],
    'ansa': [(r'.*', AnsaExtractor),],
    'citrusindustry': [(r'.*', CitrusIndustryExtractor),],
    'usda': [
        (r'/aphis/newsroom/', NewsroomExtractor),
        (r'.*', UsdaExtractor),],
    'freshplaza': [(r'.*', DailyGroupExtractor),],
    'floraldaily': [(r'.*', DailyGroupExtractor),],
    'groentennieuws': [(r'.*', DailyGroupExtractor),],
    'hortidaily': [(r'.*', DailyGroupExtractor),],
    'verticalfarmdaily': [(r'.*', DailyGroupExtractor),],
    'mmjdaily': [(r'.*', DailyGroupExtractor),],
    'oliveoiltimes': [
        (r'/(business|briefs|production|world)/', OliveoilTimesExtractor),],
    'phytoma': [(r'.*', DailyGroupExtractor),],
    'freshfruitportal': [(r'.*', FreshFruitPortalExtractor),],
    'huffingtonpost': [(r'.*', HuffingtonExtractor),],
    'ica': [(r'/noticias/', IcaExtractor),],
    'mon-viti': [(r'.*', MonVitiExtractor),],
    'nieuweoogst': [(r'/nieuws/', NieuweoogstExtractor),],
    'teatronaturale': [(r'.*', TeatroNaturaleExtractor),],
    'theguardian': [(r'.*', TheGuardianExtractor),],
    'vniikr': [(r'/news/', VniikrExtractor),],
}

# Methods

def add_types(dict_types):
    """
    Add new type domains
    dict_types: dict, {type: [domain1, domain2, ...],}
    :return: None
    """
    global TYPES
    news = dict(sum([[(i,k) for i in v] for k,v in dict_types.items()], []))
    TYPES = {**TYPES, **news}


def add_fetch_kwargs(dict_types):
    """
    Add new fetch kwargs for type domains
    dict_types: dict, {type: {kwargs1:value1, kwargs2:value2,},}
    :return: None
    """
    global FETCH_KWARGS
    news = FETCH_KWARGS.copy()
    for d_type, key_args in dict_types.items():
        news[d_type] = {**news.get(d_type, news['default']), **key_args}
    FETCH_KWARGS = news.copy()



__all__ = [
    "BaseExtractor",
    "PDFExtractor",
    "MdpiExtractor",
    "PMCExtractor",
    "NihExtractor",
    "ScienceDirectExtractor",
    "SpringerExtractor",
    "ApsNetExtractor",
    "WileyExtractor",
    "DailyGroupExtractor",
    "FreshFruitPortalExtractor",
    "HuffingtonExtractor",
    "CitrusIndustryExtractor",
    "UsdaExtractor",
    "NewsroomExtractor",
    "VniikrExtractor",
    "TheGuardianExtractor",
    "TeatroNaturaleExtractor",
    "OliveoilTimesExtractor",
    "IcaExtractor",
    "MonVitiExtractor",
    "NieuweoogstExtractor",
    "AgronewsExtractor",
    "AnsaExtractor"
]

import re
from requests import get
from requests.compat import urljoin
from requests.exceptions import InvalidURL
from requests.utils import default_headers, urlparse
from bs4 import BeautifulSoup as B
from tldextract import extract
from html import unescape
from json import loads
from pprint import pprint as cat, pformat as pcat
from datetime import datetime


from rollet import settings
from rollet.rollet import get_content
from rollet.pdfclient import GrobidClient
from rollet.utils import (
    rfinditem
)


class Extractor:
    
    def __repr__(self):
        string = "Title: {}\nFrom: {}\nFetched at: {}\nStatus: {}\nType: {}\n{} Abstract {}\n{}\n{} Full Text {}\n{}"
        return string.format(
            self.title, self.url, self.date,
            self._status, self.content_type,
            '-'*5, '-'*5, pcat(self.abstract),
            '-'*5, '-'*5, pcat(self.fulltext)
        )

    @staticmethod
    def _clean_url(url):
        return url
    
    def _init_kwargs(self, **kwargs):
        self._fulltext_kwargs, self._abstract_kwargs, self._title_kwargs, self._lang_kwargs = {}, {}, {}, {}
        for k,v in kwargs.items():
            field, key, *_ = k.split('__') + [None]
            if field == 'fulltext': self._fulltext_kwargs.update({key:v})
            elif field == 'abstract': self._abstract_kwargs.update({key:v})
            elif field == 'title': self._title_kwargs.update({key:v})
            elif field == 'lang': self._lang_kwargs.update({key:v})
    
    def _content_type(self):
        if self.url[-3:] == 'pdf': content = 'pdf'
        else:
            charset = self._header.get('Content-Type', '')
            content = re.findall('(html|pdf|json|xml)', charset)
            content = content[0] if len(content) else 'html'
        self.content_type = content

    def to_dict(self, fulltext: bool = False):
        return {
            'url': self.url,
            'status': self._status,
            'title': self.title,
            'abstract': self.abstract,
            'fulltext': self.fulltext if fulltext else None,
            'lang': self.lang,
            'content_type': self.content_type,
            'date': self.date
        }
    
    def to_list(self, *args):
        if len(args): listed = [getattr(self, arg, None) for arg in args]
        else: listed = list(self.to_dict().values())
        return listed



class BaseExtractor(Extractor):

    def __new__(cls, url, **kwargs):
        headers = default_headers()
        headers.update({
            'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
        })
        timeout = float(kwargs.get('timeout', 1))
        response = get(url, headers = headers, timeout = timeout)
        if 'pdf' in response.headers.get('Content-Type', '').lower():
            self = PDFExtractor.__new__(PDFExtractor, url, **kwargs)
            self._response = response
            self.__init__(url, **kwargs)
        else:
            self = super(BaseExtractor, cls).__new__(cls)
            self._response = response
        return self


    def __init__(self, url, **kwargs):
        """
        Create BaseExtractor instance
        url: string
        timeout: float, request timeout. Default 1 sec.
        rebond: bool, request <a> tag. Default False.
        abstract_**kwargs: **kwargs for abstract fetch
        title_**kwargs: **kwargs for title fetch
        """
        
        self.url = self._clean_url(url)
        self.domain = extract(url).domain
        self.rebond = bool(kwargs.get('rebond', False))
        self.date = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        self._scrap(**kwargs)
        self._init_kwargs(**kwargs)
    

    def _scrap(self, **kwargs):
        timeout = kwargs.pop('timeout', 1)
        headers = default_headers()
        headers.update({
            'User-Agent':
            'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
        })
        headers = kwargs.pop('headers', headers)

        if not hasattr(self, '_response'):
            with get(self.url, timeout = timeout, headers = headers) as response:
                response = response
        else:
            response = self._response
        self._status = response.status_code
        self._header = response.headers
        self._content_type()
        if self.content_type == 'html':
            self._page = B(response.text, 'html.parser')
        else:
            self._page = B()


    def _get_content(self, tag, **kwargs):
        """
        Get content from element Tag
        tag: bs4.element.Tag
        script_keys: List, of keys if tag is script
        attr: str, key attribute if tag is neither a meta nor a script
        fields: List, of fields to concat if rebond
        :return: content
        """
        fields = kwargs.get('fields', ['fulltext'])
        if tag.name == 'meta':
            content = tag.attrs.get('content', [])
        elif tag.name == 'a' and tag.attrs.get('href') and self.rebond:
            self._presave = True
            raw_content = get_content(urljoin(self.url, tag.attrs['href']), rebond = False)
            content = '. '.join(raw_content.get(field, '') if raw_content.get(field, '') else '' for field in fields )
        elif tag.name == 'script':
            keys = kwargs.get('script_keys', [])
            try:
                serie = loads(tag.content[0])
            except:
                content = list()
            else:
                content = [rfinditem(serie, key) for key in keys]
        elif kwargs.get('attr'):
            content = tag.attrs.get(kwargs.get('attr'))
        else:
            content = tag.get_text(separator = ' ').strip().replace('\n', ' ')

        content = content[0] if isinstance(content, list) and len(content) else content
        content = unescape(content) if isinstance(content, str) else content
        return content
    

    def __repr__(self):
        string = "Title: {}\nFrom: {}\nFetched at: {}\nStatus: {}\nType: {}\n{} Abstract {}\n{}\n{} Full Text {}\n{}"
        return string.format(
            self.title, self.url, self.date,
            self._status, self.content_type,
            '-'*5, '-'*5, pcat(self.abstract),
            '-'*5, '-'*5, pcat(self.fulltext)
        )


    def fetch(self, selectors, which='first', **kwargs):
        content = list()
        arg_w = ('first', 'min', 'max')
        if which not in arg_w:
            raise ValueError(f'which should be one of {arg_w}')
        for s in selectors:
            contents_ = list()
            tags = self._page.select(s)
            if len(tags): contents_ = [self._get_content(tag, **kwargs) for tag in tags]
            if len(contents_) and which == 'first': 
                content = [contents_[0]]
                break
            else:
                try: content += ['. '.join(set(contents_))]
                except: pass
        content = content if len(content) else [None]
        if which == 'max': content = max(content, key=lambda x: len(str(x)))
        else: content = min(content, key=lambda x: len(str(x)))
        return content
    
    @property
    def title(self):
        title = None
        if self.content_type == 'html':
            title = self.fetch(settings.TITLE, **self._title_kwargs)
        return title
    
    @property
    def abstract(self):
        abstract = None
        if self.content_type == 'html':
            abstract = self.fetch(settings.ABSTRACT, **self._abstract_kwargs)
        return abstract
    
    @property
    def fulltext(self):
        fulltext = getattr(self, '_fulltext', None)
        if not fulltext and self.content_type == 'html':
            fulltext = self.fetch(settings.FULLTEXT, fields = ['fulltext'], **self._fulltext_kwargs)
            self._fulltext = fulltext if getattr(self, '_presave', False) else None
        return fulltext
    
    @property
    def lang(self):
        lang = None
        if self.content_type == 'html':
            lang = self._page.html.get('lang', None)
        return lang



class PDFExtractor(Extractor):
    def __init__(self, url, **kwargs):
        """
        Create PDFExtractor instance
        url: string
        timeout: float, request timeout. Default 1 sec.
        rebond: bool, request <a> tag. Default False.
        abstract_**kwargs: **kwargs for abstract fetch
        title_**kwargs: **kwargs for title fetch
        """
        
        self.url = self._clean_url(url)
        self.domain = extract(url).domain
        self.date = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        self._scrap(**kwargs)
        
    def _scrap(self, **kwargs):
        timeout = float(kwargs.pop('timeout', 1))
        headers = default_headers()
        headers.update({
            'User-Agent':
            'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
        })
        headers = kwargs.pop('headers', headers)

        if not hasattr(self, '_response'):
            with get(self.url, timeout = timeout, headers = headers) as response:
                response = response
        else:
            response = self._response
        self._status = response.status_code
        self._header = response.headers
        self._content_type()
        if self.content_type == 'pdf':
            try: self._page = self._pdf_parse(response.content)
            except: self._page = B()
        else:
            self._page = B()
        
        
    @staticmethod
    def _pdf_parse(pdf_file):
        grobid_client = GrobidClient()
        req = grobid_client.serve('processFulltextDocument', pdf_file)
        if req.status_code == 200:
            return B(req.text, 'html.parser')
        else: return B()
    
    @property
    def title(self):
        titles = self._page.select('title[level=a i][type=main i]')
        try: title = titles[0].get_text(' ', True)
        except: title = None
        return title
    
    @property
    def abstract(self):
        if not hasattr(self._page, 'body'): return None
        abstracts = self._page.select(':is(abstract,details,summary)')
        abstracts += [self._page.body.find('div')] if self._page.body and self._page.body.find('div') else []
        try:abstract = abstracts[0].get_text(' ', True)
        except: abstract = None
        return abstract
    
    @property
    def fulltext(self):
        try:
            fulltext = '. '.join(
                map(lambda x: x.get_text(' ', True),
                    self._page.body.findAll('div', recursive = False)))
        except: fulltext = None
        return fulltext
    
    @property
    def lang(self):
        try: lang = self._page.select(':is(teiheader,text)')[-1].get('xml:lang', 'en')
        except: lang = None
        return lang



class BaseContentExtractor(BaseExtractor):
    """
    Extractor for content in list
    """
    
    def _content(self):
        return list()
    
    @property
    def abstract(self):
        abstract = None
        try: abstract = self._content()[0].get_text(' ', True)
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        fulltext = None
        try: fulltext = '. '.join(map(lambda x:x.get_text(' ', True), self._content()[1:]))
        except: fulltext = super().fulltext
        return fulltext



class MdpiExtractor(BaseExtractor):
    """
    Extractor for base mdpi.com domain
    """
    
    @staticmethod
    def _clean_url(url):
        parsed = urlparse(url)
        path = re.sub(r'/(htm|pdf|xml|scifeed_display|notes|reprints|s[0-9]|review_report)?/?$', '', parsed.path) + '/htm'
        return f'{urljoin(url, path)}?{parsed.query}'
    
    @property
    def abstract(self):
        try: abstract = self._page.find(class_='art-abstract').get_text(separator = ' ').strip()
        except: abstract = super().abstract
        return abstract

    @property
    def title(self):
        try:
            title = self._page.find(class_='title', attrs={
                'itemprop': 'name'
            }).get_text(separator = ' ').strip()
        except: title = super().title
        return title
    
    @property
    def fulltext(self):
        self._fulltext = '. '.join(map(
            lambda x:x.get_text(' ', True),
            self._page.select('div[class="html-body"] section')))
        fulltext = super().fulltext if not len(self._fulltext) else self._fulltext
        return fulltext



class PMCExtractor(BaseContentExtractor):
    """
    Extractor for PubMed Central articles
    """
    
    def _content(self):
        return self._page.select('div[class="tsec sec" i]')
    
    @property
    def title(self):
        title = None
        try: title = self._page.find('h1', class_='content-title').get_text(' ', True)
        except: title = super().title
        return title



class NihExtractor(BaseExtractor):
    """
    Extractor for base nih.gov domain
    """
    
    def __new__(cls, url, **kwargs):
        """
        Redirect to PMC version if exist
        """
        headers = default_headers()
        headers.update({
            'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
        })
        timeout = float(kwargs.get('timeout', 1))
        response = get(url, headers = headers, timeout = timeout)
        links = B(response.text, 'html.parser').select('a[class*="pmc"][data-ga-category*="full_text"]')
        if len(links):
            pmc_url = list({u.attrs['href'] for u in links if u.attrs.get('href')})[0]
            pmc_url = urljoin(url, pmc_url)
            return PMCExtractor(pmc_url, **kwargs)
        self = super(NihExtractor, cls).__new__(cls, url, **kwargs)
        self._response = response
        return self
    
    @property
    def abstract(self):
        try: abstract = self._page.find(id='enc-abstract').get_text(separator = ' ').strip()
        except: abstract = None
        finally: abstract = super().abstract if not abstract else abstract
        return abstract

    @property
    def title(self):
        try: title = self._page.find('h1', class_='heading-title').get_text(separator = ' ').strip()
        except: title = None
        finally: title = super().title if not title else title
        return title
    
    @property
    def fulltext(self):
        if getattr(self, '_fulltext', None): return self._fulltext
        full_url = list({
            u.attrs['href']
            for u in self._page.select('a[data-ga-category*="full_text" i]')
            if u.attrs.get('href')})
        try: _fulltext = get_content(full_url[-1], rebond = False)['fulltext']
        except: _fulltext = None
        if _fulltext:
            self._presave = True
            self._fulltext = _fulltext
        return super().fulltext



class ScienceDirectExtractor(BaseExtractor):
    """
    Extractor for base sciencedirect.com domain
    """

    @property
    def abstract(self):
        try:
            abstract = self._page.find(id='abstracts',
                                   class_='Abstracts').get_text(' ').strip()
        except: abstract = None
        abstract = super().abstract if not abstract else abstract
        return abstract

    @property
    def title(self):
        try: title = self._page.find('span', class_='title-text').get_text(' ').strip()
        except: title = None
        title = super().title if not title else title
        return title



class SpringerExtractor(BaseExtractor):
    """
    Extractor for base springer.com domain
    """
    
    @property
    def title(self):
        title = self._page.find('h1', class_ = 'c-article-title')
        title = title.get_text(' ').strip() if title else super().title
        return title
    
    @property
    def abstract(self):
        try:
            abstract = list(set(map(
                lambda x:x.get_text(' ', True),
                self._page.select(
                    'div[id^="Abs"][class*="c-article-section__content"]'
                ))))[0]
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try:
            self._fulltext = '. '.join(set(map(
                lambda x:x.get_text(' ', True),
                self._page.select(
                    'div[id^="Sec"][class="c-article-section"]'
                ))))
        except: self._fulltext = None
        return super().fulltext



class ApsNetExtractor(BaseExtractor):
    
    @staticmethod
    def _clean_url(url):
        return re.sub(
            r'(/doi/)[a-zA-Z]*/?([0-9])',
            r'\1full/\2', url)
    
    @property
    def title(self):
        try: title = self._page.find('h1', class_ = 'citation__title').get_text(' ', True)
        except: title = None
        finally: title = title if title else super().title
        return title
        
    @property
    def abstract(self):
        try: abstract = self._page.find('div', class_ = 'hlFld-Abstract').get_text(' ', True)
        except: abstract = None
        finally: abstract = abstract if abstract else super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try: self._fulltext = self._page.find('div', class_ = 'hlFld-Fulltext').get_text(' ', True)
        except: self._fulltext = None
        return super().fulltext



class WileyExtractor(BaseExtractor):
    
    @staticmethod
    def _clean_url(url):
        return re.sub(
            r'(/doi/)[a-zA-Z]*/?([0-9])',
            r'\1full/\2', url)
        
    @property
    def abstract(self):
        try: abstract = self._page.find(
            class_ = 'article-section article-section__abstract').get_text(' ', True)
        except: abstract = None
        finally: abstract = abstract if abstract else super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try: self._fulltext = self._page.find(
            class_ = 'article-section article-section__full').get_text(' ', True)
        except: self._fulltext = None
        return super().fulltext



class DailyGroupExtractor(BaseContentExtractor):
    """
    Extractor for Daily Group
    """
    
    def _content(self):
        return self._page.select('div[itemprop="articleBody" i] p')
    
    @property
    def title(self):
        title = None
        try: title = self._page.find('h1', itemprop = 'name headline').get_text(' ', True)
        except: title = super().title
        return title



class FreshFruitPortalExtractor(BaseContentExtractor):
    """
    Extractor for Daily Group
    """
    
    def _content(self):
        return self._page.select('div[class="post-content" i] p')
    
    @property
    def title(self):
        title = None
        try: title = self._page.find(class_ = 'post-title').get_text(' ', True)
        except: title = super().title
        return title



class HuffingtonExtractor(BaseContentExtractor):
    """
    Extractor for HuffingtonPost
    """
    
    def _content(self):
        return self._page.select('.post-contents .content-list-component.text > p')
    
    @property
    def title(self):
        title = None
        try: title = self._page.find(class_ = 'headline__title').get_text(' ', True)
        except: title = super().title
        return title



class CitrusIndustryExtractor(BaseContentExtractor):
    """
    Extractor for Citrusindustry
    """
    
    def _content(self):
        return self._page.select('.entry-content.content > p')
    
    @property
    def title(self):
        try: title = self._page.find(class_ = 'entry-title').get_text(' ', True)
        except: title = super().title
        return title



class UsdaExtractor(BaseContentExtractor):
    """
    Extractor for Usda
    """
    
    def _content(self):
        return self._page.select('#MainContent :is(p,ul,ol)')
    
    @property
    def title(self):
        try: title = self._page.find(class_ = 'contentTitle').get_text(' ', True)
        except: title = super().title
        return title



class NewsroomExtractor(UsdaExtractor, BaseContentExtractor):
    """
    Extractor for Aphis.usda Newsroom
    """
    
    def _content(self):
        try: content = self._page.find(class_ = 'contentTitle').parent.select(':is(p,ul,ol)')
        except: content = super()._content()
        return content



class VniikrExtractor(BaseExtractor):
    """
    Extractor for vniikr
    """
    def _content(self):
        try:
            contents = [
                x.strip() for x in self._page.select_one(
                    '.article-content.container').contents
                if isinstance(x, str) and x != '\n'
            ]
        except: contents = list()
        return contents

    @property
    def title(self):
        try:
            title = self._page.select_one(
                '.article-content.container h2').get_text(' ', True)
        except: title = super().title
        return title
    
    @property
    def abstract(self):
        try: abstract = self._content()[0]
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try: fulltext = ' '.join(self._content()[1:])
        except: fulltext = super().fulltext
        return fulltext



class TheGuardianExtractor(BaseExtractor):
    """
    Extractor for the Guardian
    """

    @property
    def title(self):
        try:
            title = self._page.select_one(
                'article h1').get_text(' ', True)
        except: title = super().title
        return title
    
    @property
    def abstract(self):
        try: abstract = self._page.select_one('article p').get_text(' ', True)
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try:
            fulltext = ' '.join(
                map(lambda x: x.get_text(' ', True),
                    self._page.select('#maincontent div[class^=article i] :is(p,ul,ol)')))
        except: fulltext = super().fulltext
        return fulltext



class TeatroNaturaleExtractor(BaseExtractor):
    """
    Extractor for teatronaturale
    """

    @property
    def title(self):
        try:
            title = self._page.select_one(
                'main h1').get_text(' ', True)
        except: title = super().title
        return title
    
    @property
    def abstract(self):
        try: abstract = self._page.select_one('.sommario p').get_text(' ', True)
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try:
            fulltext = ' '.join(
                map(lambda x: x.get_text(' ', True),
                    self._page.select('.contenuto :is(p,ul,ol)')))
        except: fulltext = super().fulltext
        return fulltext



class OliveoilTimesExtractor(BaseExtractor):
    """
    Extractor for oliveoiltimes
    """

    @property
    def title(self):
        try:
            title = self._page.select_one(
                'div[id*=cover-news i] h2').get_text(' ', True)
        except: title = super().title
        return title
    
    @property
    def abstract(self):
        try: abstract = self._page.select_one('div[id*=cover-news i] .bigger-p').get_text(' ', True)
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try:
            fulltext = ' '.join(
                map(lambda x: x.get_text(' ', True),
                    self._page.select('#single-post-content :is(p,ul,ol)')))
        except: fulltext = super().fulltext
        return fulltext



class IcaExtractor(BaseExtractor):
    """
    Extractor for ica.gov
    """

    @property
    def title(self):
        try:
            title = self._page.select_one(
                'article h2').get_text(' ', True)
        except: title = super().title
        return title
    
    @property
    def abstract(self):
        try: abstract = self._page.select_one(
            'article :is(p,ul,ol)').get_text(' ', True)
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try:
            fulltext = self._page.select_one('article').find(
            'hr').fetchNextSiblings('div')[0].get_text(' ', True)
        except: fulltext = super().fulltext
        return fulltext



class MonVitiExtractor(BaseContentExtractor):
    """
    Extractor for mon-viti
    """
    
    def _content(self):
        return self._page.select('article section :is(p,ul,ol,blockquote)')
    
    @property
    def title(self):
        try:
            title = self._page.select_one(
                'article h1[class*=title i]').get_text(' ', True)
        except: title = super().title
        return title



class NieuweoogstExtractor(BaseExtractor):
    """
    Extractor for nieuweoogst
    """
    
    @property
    def title(self):
        try:
            title = self._page.select_one(
                '.content h1').get_text(' ', True)
        except: title = super().title
        return title
    
    @property
    def abstract(self):
        try: abstract = self._page.select_one(
            '.content p.intro').get_text(' ', True)
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try:
            fulltext = ' '.join(
                map(lambda x: x.get_text(' ', True),
                    _page.select('.content [id*=lightgallery i] :is(p,ul,ol)')))
        except: fulltext = super().fulltext
        return fulltext



class AgronewsExtractor(BaseExtractor):
    """
    Extractor for agronews
    """
    
    @property
    def title(self):
        try:
            title = self._page.select_one(
                'main section [class*=titlewrp i] h1').get_text(' ', True)
        except: title = super().title
        return title
    
    @property
    def abstract(self):
        try: abstract = self._page.select_one(
            'main section [class*=titlewrp i] div[class*=txt i]').get_text(' ', True)
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try:
            fulltext = ' '.join(
                map(lambda x: x.get_text(' ', True),
                    _page.select('article [class*=desc i] :is(p,ul,ol,blockquote)')))
        except: fulltext = super().fulltext
        return fulltext



class AnsaExtractor(BaseExtractor):
    """
    Extractor for ansa
    """
    
    @property
    def title(self):
        try:
            title = self._page.select_one(
                'header h1[class*=title i]').get_text(' ', True)
        except: title = super().title
        return title
    
    @property
    def abstract(self):
        try: abstract = self._page.select_one(
            'header h2[class*=news i]').get_text(' ', True)
        except: abstract = super().abstract
        return abstract
    
    @property
    def fulltext(self):
        try:
            fulltext = ' '.join(
                map(lambda x: x.get_text(' ', True),
                    _page.select('div[class*=news-txt i] :is(p,ul,ol)')))
        except: fulltext = super().fulltext
        return fulltext




__all__ = [
    "GrobidClient",
    "grobid_service"
]

from typing import List, Dict, Union
from requests.compat import urljoin
from copy import deepcopy
import json, requests, os

class ApiClient(object):
    accept_type = 'application/xml'
    api_base = None

    def __init__(
            self,
            base_url: str,
            username: str = None,
            api_key: str = None,
            status_endpoint: str = None,
            timeout: int = 60
    ):
        """
        Initialise client.
        base_url: The base URL to the service being used.
        username: The username to authenticate with.
        api_key: The API key to authenticate with.
        timeout: Maximum time before timing out.
        """
        self.base_url = base_url
        self.username = username
        self.api_key = api_key
        self.status_endpoint = urljoin(self.base_url, status_endpoint)
        self.timeout = timeout

    @staticmethod
    def encode(request, data):
        """ Add request content data to request body, set Content-type header.
        Should be overridden by subclasses if not using JSON encoding.
        Args:
            request (HTTPRequest): The request object.
            data (dict, None): Data to be encoded.
        Returns:
            HTTPRequest: The request object.
        """
        if data is None:
            return request

        request.add_header('Content-Type', 'application/json')
        request.data = json.dumps(data)

        return request

    @staticmethod
    def decode(response):
        """ Decode the returned data in the response.
        Should be overridden by subclasses if something else than JSON is
        expected.
        Args:
            response (HTTPResponse): The response object.
        Returns:
            dict or None.
        """
        try:
            return response.json()
        except ValueError as e:
            return e.message

    def get_credentials(self) -> Dict:
        """ Returns parameters to be added to authenticate the request.
        This lives on its own to make it easier to re-implement it if needed.
        Returns:
            dict: A dictionary containing the credentials.
        """
        return {"username": self.username, "api_key": self.api_key}

    def call_api(
            self,
            method: str,
            url: str,
            headers: Dict[str, str] = None,
            params: Dict[str, str] = None,
            data: Dict[str, str] = None,
            files: Dict[str, str] = None,
            timeout: int = None,
    ):
        """
        This returns object containing data, with error details if applicable.
        method: The HTTP method to use.
        url: Resource location relative to the base URL.
        headers: Extra request headers to set.
        params: Query-string parameters.
        data: Request body for POST or PUT requests.
        files: Files to be passed to the request.
        timeout: Maximum time before timing out.

        :return: ResultParser or ErrorParser.
        """
        headers = deepcopy(headers) or {}
        headers['Accept'] = self.accept_type
        params = deepcopy(params) or {}
        data = data or {}
        files = files or {}
        r = requests.request(
            method,
            url,
            headers=headers,
            params=params,
            files=files,
            data=data,
            timeout=timeout,
        )

        return r, r.status_code

    def get(self,
        url: str,
        params: Dict[str, str] = None,
        **kwargs
    ):
        """ 
        Call the API with a  GET request.
        url: Resource location relative to the base URL.
        params: Query-string parameters.
        :return: ResultParser or ErrorParser.
        """
        return self.call_api(
            "GET",
            url,
            params=params,
            **kwargs
        )

    def delete(self, url, params=None, **kwargs):
        """ Call the API with a DELETE request.
        Args:
            url (str): Resource location relative to the base URL.
            params (dict or None): Query-string parameters.
        Returns:
            ResultParser or ErrorParser.
        """
        return self.call_api(
            "DELETE",
            url,
            params=params,
            **kwargs
        )

    def put(self, url, params=None, data=None, files=None, **kwargs):
        """ Call the API with a PUT request.
        Args:
            url (str): Resource location relative to the base URL.
            params (dict or None): Query-string parameters.
            data (dict or None): Request body contents.
            files (dict or None: Files to be passed to the request.
        Returns:
            An instance of ResultParser or ErrorParser.
        """
        return self.call_api(
            "PUT",
            url,
            params=params,
            data=data,
            files=files,
            **kwargs
        )

    def post(self, url, params=None, data=None, files=None, **kwargs):
        """ Call the API with a POST request.
        Args:
            url (str): Resource location relative to the base URL.
            params (dict or None): Query-string parameters.
            data (dict or None): Request body contents.
            files (dict or None: Files to be passed to the request.
        Returns:
            An instance of ResultParser or ErrorParser.
        """
        return self.call_api(
            method="POST",
            url=url,
            params=params,
            data=data,
            files=files,
            **kwargs
        )

    def service_status(self, **kwargs):
        """ Call the API to get the status of the service.
        Returns:
            An instance of ResultParser or ErrorParser.
        """
        return self.call_api(
            'GET',
            self.status_endpoint,
            params={'format': 'json'},
            **kwargs
        )


class GrobidClient(ApiClient):

    def __init__(self,
        host: str = None,
        port: int = None
    ):
        if not host: host = str(os.getenv('GROBID_HOST', 'localhost'))
        if not port: port = int(os.getenv('GROBID_PORT', 8070))

        if not host.startswith("http"):
            host = "http://" + host
        self.host = host
        self.port = port
        self.url = f"{self.host}:{self.port}"

    def test_alive(self):
        url = f"{self.url}/api/isalive"
        rsp = requests.get(url)
        return rsp.status_code == 200

    def serve(
        self,
        service: str,
        pdf_file,
        generateIDs: int = 1,
        consolidate_header: int = 0,
        consolidate_citations: int = 0,
        teiCoordinates: List = ["persName", "figure", "ref", "biblStruct", "formula"]
    ):

        if isinstance(pdf_file, str):
            with open(pdf_file, 'rb') as fp:
                pdf_file = fp
        files = {'input': pdf_file}
        url = f"{self.url}/api/{service}"

        the_data = {
            "generateIDs": generateIDs,
            "consolidateHeader": consolidate_header,
            "consolidateCitations": consolidate_citations,
            "teiCoordinates": teiCoordinates
        }

        rsp, _ = self.post(
            url = url,
            files = files,
            data = the_data,
            headers = {'Accept': 'text/plain'}
        )

        return rsp
    

def grobid_service(
    host: str = 'localhost',
    port: Union[str, int] = '8070'
):
    os.environ.update(**{'GROBID_HOST': str(host), 'GROBID_PORT': str(port)})
